//
//  SynthVCO.m
//
//  Created by 成田 拓郎 on 10/04/30.
//  Copyright 2010 lovelyshark. All rights reserved.
//

#import "SynthVCO.h"


@implementation SynthVCO

//ノコギリ波
+ (float*) waveSawTooth
{
	static float wave[SAMPLE_RATE];

	static BOOL flagWave = YES;
	if (flagWave) {
		float range = 0.0f;
		for (int i=0; i<SAMPLE_RATE; ++i) {
			float w = 0.0f;
			for (int n = 1; n<=21; ++n) {
				w += (1.0f / n) * sinf(2.0f * M_PI * n * i / SAMPLE_RATE);
			}
			wave[i] = w;
			if (range < fabsf(w)) range = fabsf(w);
		}
		for (int i=0; i<SAMPLE_RATE; ++i)
			wave[i] /= range;

		flagWave = NO;
	}

	return wave;
}

//矩形波
+ (float*) waveSquare
{
	static float wave[SAMPLE_RATE];
	
	static BOOL flagWave = YES;
	if (flagWave) {
		float range = 0.0f;
		for (int i=0; i<SAMPLE_RATE; ++i) {
			float w = 0.0f;
			for (int n = 1; n<=21; n+=2) {
				w += (1.0f / n) * sinf(2.0f * M_PI * n * i / SAMPLE_RATE);
			}
			wave[i] = w;
			if (range < fabsf(w)) range = fabsf(w);
		}
		//振幅を1以下にする
		for (int i=0; i<SAMPLE_RATE; ++i)
			wave[i] /= range;

		flagWave = NO;
	}
	
	return wave;
}

//三角波
+ (float*) waveTriangle
{
	static float wave[SAMPLE_RATE];
	
	static BOOL flagWave = YES;
	if (flagWave) {
		float range = 0.0f;
		for (int i=0; i<SAMPLE_RATE; ++i) {
			float w = 0.0f;
			for (int n = 1; n<=21; ++n) {
				w += (8.0f / (M_PI * M_PI)) * sinf(n * M_PI / 2.0f) * sinf(2.0f * M_PI * n * i / SAMPLE_RATE) / (n * n);
			}
			wave[i] = w;
			if (range < fabsf(w)) range = fabsf(w);
		}
		for (int i=0; i<SAMPLE_RATE; ++i)
			wave[i] /= range;

		flagWave = NO;
	}
	
	return wave;
}

//パルス波
- (float*) wavePulse
{
	static float wave[SAMPLE_RATE];
	
	static BOOL flagWave = YES;
	static float pw = 0.0f;
	if (flagWave || (pw != pulseWidth)) {
		int pulse_delta = (int)(SAMPLE_RATE * pulseWidth + 0.5f);
		for (int i=0; i<SAMPLE_RATE; ++i) {
			wave[i] = 0.0f;
		}
		for (int i=0; i<SAMPLE_RATE; ++i) {
			float w = 0.0f;
			for (int n = 1; n<=21; ++n) {
				float p = 2.0f * M_PI * n * i / SAMPLE_RATE;
				w += (1.0f / n) * sinf(p);
			}
			wave[i] += w;
			wave[(i + pulse_delta) % SAMPLE_RATE] -= w;
		}
		float range = 0.0f;
		for (int i=0; i<SAMPLE_RATE; ++i){
			if (range < fabsf(wave[i]))
				range = fabsf(wave[i]);
		}
		for (int i=0; i<SAMPLE_RATE; ++i)
			wave[i] /= range;

		flagWave = NO;
		pw = pulseWidth;
	}

	return wave;
}

//サイン波
+ (float*) waveSin
{
	static float wave[SAMPLE_RATE];
	
	static BOOL flagWave = YES;
	if (flagWave) {
		for (int i=0; i<SAMPLE_RATE; ++i) {
			wave[i] = sinf((M_PI * 2.0f * i) / SAMPLE_RATE);
		}
		
		flagWave = NO;
	}
	
	return wave;
}

//ホワイトノイズ
+ (float*) waveWhiteNoise
{
	static float wave[SAMPLE_RATE];
	
	static BOOL flagWave = YES;
	if (flagWave) {
		for (int i=0; i<SAMPLE_RATE; ++i) {
			wave[i] = [SynthObject AWGN];
		}

		flagWave = NO;
	}
	
	return wave;
}

//選択波形
- (float*) waveBase
{
	switch (waveType) {
		case VCO_SAW_TOOTH_WAVE:
			return [SynthVCO waveSawTooth];
		case VCO_SQUARE_WAVE:
			return [SynthVCO waveSquare];
		case VCO_TRIANGLE_WAVE:
			return [SynthVCO waveTriangle];
		case VCO_PULSE_WAVE:
			return [self wavePulse];
		case VCO_SIN_WAVE:
			return [SynthVCO waveSin];
		case VCO_WHITE_NOISE:
		default:
			return [SynthVCO waveWhiteNoise];
	}
}

//ノート番号→周波数
+ (float) note2Frequency:(float)note_number
{
	return 440.0f * powf(2.0f, (note_number - 69.0f) / 12.0f);
}

//プロパティ

- (int) waveType
{
	return waveType;
}
- (void) setWaveType:(int)type_number
{
	waveType = type_number;
	phase = (2.0f * M_PI * random()) / INT32_MAX;
}

- (float) pulseWidth
{
	return pulseWidth;
}
- (void) setPulseWidth:(float)pulse_width
{
	if (pulse_width <= 0.0f)
		pulseWidth = 0.0f;
	else if (pulse_width >= 1.0f)
		pulseWidth = 1.0f;
	else
		pulseWidth = pulse_width;
}

@synthesize frequency;
@synthesize phase;

//このオブジェクトの演算関数
- (void) synthFunc:(int)length
{
	float freq_param  = controlLevel * frequency * VCO_LFO;
	
	float delta_time  = 1.0f / [SynthObject outputSampleRate:0.0f];			//オーディオ出力１サンプルの時間(s)
	float step = SAMPLE_RATE * phase / (2.0f * M_PI);	//VCO波形の開始位相(ステップ)
	step = fmodf(step, SAMPLE_RATE);
	
	float *wave_base = [self waveBase];	//基本波形

	float *waveout_dest = outputBuffer;
	float *waveout_end  = waveout_dest + length;
	float *waveout_src = (inputControlObject) ? inputControlObject.outputBuffer : [SynthObject dummyBuffer0];

	float phase_sum = 0.0f;
	float freq, step_delta_phase, delta_phase;
	int lfo_count = 0;
	while (waveout_dest != waveout_end) {
		int wi = (int)(step + 0.5f);
		if (SAMPLE_RATE <= wi) {
			step -= SAMPLE_RATE;
			wi   -= SAMPLE_RATE;
		}
		*waveout_dest = wave_base[wi];
		++waveout_dest;

		// LFOは250サンプルに1回計算し直す
		if (lfo_count == 0) {
			lfo_count = 249;

			freq = frequency + (*waveout_src) * freq_param;	//LFO変調済み発音周波数
//			delta_phase = 2.0f * M_PI * freq * delta_time;	//１サンプルで進むVCO波形の位相(ラジアン)
//			step_delta_phase = SAMPLE_RATE * delta_phase / (2.0f * M_PI);	//１サンプルで進むVCO波形の位相(ステップ)
			delta_phase = freq * delta_time;	//１サンプルで進むVCO波形の位相(ラジアン)
			step_delta_phase = SAMPLE_RATE * delta_phase;	//１サンプルで進むVCO波形の位相(ステップ)
		}
		else
			--lfo_count;

		++waveout_src;
		step      += step_delta_phase;
		phase_sum += delta_phase;
	}
//	phase = fmodf(phase + phase_sum, 2.0f * M_PI);
	phase = fmodf(phase + (phase_sum * 2.0f * M_PI), 2.0f * M_PI);
}

- (id) init
{
	if (self = [super init])
	{
		self.waveType   = VCO_TRIANGLE_WAVE;
		self.pulseWidth = 0.5f;
		self.frequency  = 440.0f;

		[SynthVCO waveSawTooth];
		[SynthVCO waveSquare];
		[SynthVCO waveTriangle];
		[SynthVCO waveSin];
		[SynthVCO waveWhiteNoise];
	}

	return self;
}

- (void) dealloc
{
	
	[super dealloc];
}


@end
